@@ -9,10 +9,12 @@ cmdoret, 20200404
 import sys
 import itertools as it
 from typing import Iterable, Optional, Union, Tuple
+from functools import reduce
 import numpy as np
 import pandas as pd
 import scipy.sparse as sp
 import cooler
+import multiprocessing as mp
 import chromosight.utils.preprocessing as cup
 import chromosight.utils.detection as cud
 import chromosight.utils.io as cio
@@ -20,7 +22,7 @@ import chromosight.kernels as ck
 import pareidolia.io as pai
 import pareidolia.preprocess as pap
 import pareidolia.detection as pad
-import multiprocessing as mp
+import pareidolia.stats as pas
 
 
 def get_min_contacts(
@@ -37,9 +39,7 @@ def get_min_contacts(
         if region is None:
             contacts[i] = clr.info["sum"]
         else:
-            contacts[i] = (
-                clr.matrix(balance=False, sparse=True).fetch(region).sum()
-            )
+            contacts[i] = clr.matrix(balance=False, sparse=True).fetch(region).sum()
     # Return the minimum coverage value (in number of contacts)
     return int(min(contacts))
 
@@ -86,20 +86,20 @@ def preprocess_hic(
 
 def coords_to_bins(clr: cooler.Cooler, coords: pd.DataFrame) -> np.ndarray:
     """
-        Converts genomic coordinates to a list of bin ids based on the whole genome
-        contact map.
+    Converts genomic coordinates to a list of bin ids based on the whole genome
+    contact map.
 
-        Parameters
-        ----------
-        coords : pandas.DataFrame
-            Table of genomic coordinates, with columns chrom, pos.
+    Parameters
+    ----------
+    coords : pandas.DataFrame
+        Table of genomic coordinates, with columns chrom, pos.
 
-        Returns
-        -------
-        numpy.array of ints :
-            Indices in the whole genome matrix contact map.
+    Returns
+    -------
+    numpy.array of ints :
+        Indices in the whole genome matrix contact map.
 
-        """
+    """
     coords.pos = (coords.pos // clr.binsize) * clr.binsize
     # Coordinates are merged with bins, both indices are kept in memory so that
     # the indices of matching bins can be returned in the order of the input
@@ -121,17 +121,119 @@ def coords_to_bins(clr: cooler.Cooler, coords: pd.DataFrame) -> np.ndarray:
     return idx
 
 
+def make_density_filter(
+    mats: Iterable[sp.csr_matrix],
+    density_thresh: float = 0.10,
+    win_size=3,
+    sym_upper=False,
+) -> sp.csr_matrix:
+    """Given a list of sparse matrices, generate a "density filter". This
+    new sparse matrix is a boolean mask where values indicate whether the
+    proportion of nonzero pixels in the neighbourhood of diameter win_size
+    is above the input threshold in all input matrices.
+
+    Parameters
+    ----------
+    mats : Iterable of scipy.sparse.csr_matrix
+        The matrices to be combined into a filter.
+    density_thresh : float
+        The required proportion of nonzero pixels in the neighbourhood pass
+        the filter.
+    win_size : int
+        The diameter of the neighbourhood in which to compute the proportion
+        of nonzero pixels.
+    sym_upper : bool
+        Whether the matrix is symmetric upper. In this case, computations are
+        performed in the upper triangle.
+
+    Returns
+    -------
+    scipy.sparse.csr_matrix of bools:
+        The sparse boolean mask representing the density filter. Values are
+        True where all input matrices passed the threshold.
+    """
+
+    # Filter out regions where contacts were too sparse in all contact matrices
+    # Pixel density is first converted to a boolean matrix (pass / fail)
+    filters = map(
+        lambda m: pad.get_win_density(m, win_size=win_size, sym_upper=sym_upper)
+        > density_thresh,
+        mats,
+    )
+    # matrices are reduced through an element-wise multiplication acting as an AND
+    # filter: AND(AND(AND(m1, m2), m3), m4)
+    inter_filter = reduce(lambda x, y: x.multiply(y), filters)
+    return inter_filter
+
+
+def _ttest_matrix(samples: pd.DataFrame, control: str) -> Tuple[sp.csr_matrix, float]:
+    """
+    Performs pixel-wise t-test comparisons between conditions to detect differential
+    interactions.
+    """
+    # Compute background for each condition
+    arr_control = np.dstack([m.data for m in samples.mat[samples.cond == control]])[
+        0, :, :
+    ]
+    arr_alt = np.dstack([m.data for m in samples.mat[samples.cond != control]])[0, :, :]
+    diff = samples["mat"][0]
+    diff.data = pas.vec_ttest(arr_control, arr_alt)
+    # The threshold is the t-value corresponding to p=0.05
+    # thresh = pas.pval_to_tval(
+    #    1 - percentile_thresh / 100,
+    #    arr_control.shape[1] + arr_alt.shape[1],
+    # )
+    return diff
+
+
+def _median_bg_subtraction(
+    samples: pd.DataFrame, control: str
+) -> Tuple[sp.csr_matrix, float]:
+    """
+    Performs the median background subtraction to extract differential signal
+    from multiple Hi-C matrix.
+    """
+    # Compute background for each condition
+    backgrounds = samples.groupby("cond")["mat"].apply(
+        lambda g: pad.median_bg(g.reset_index(drop=True))
+    )
+    # Compute sse for each condition
+    sse = samples.groupby("cond")["mat"].apply(
+        lambda g: pad.get_sse_mat(g.reset_index(drop=True))
+    )
+    conditions = np.unique(samples.cond)
+    # Compute difference between conditions and signal
+    # to noise ratio
+    snr = np.zeros(sse[0].data.shape)
+    diff = sp.csr_matrix(sse[0].shape)
+    for c in conditions:
+        snr += backgrounds[c].data / np.sqrt(sse[c].data)
+        if c != control:
+            # Break ties to preserve sparsity (do not introduce 0s)
+            ties = backgrounds[c].data == backgrounds[control].data
+            backgrounds[c].data[ties] += 1e-06
+            diff += backgrounds[c] - backgrounds[control]
+    snr /= len(conditions)
+    # Use average difference to first background as change metric
+    diff.data /= len(conditions) - 1
+    # Threshold data on background / sse value
+    diff.data[snr < 1.0] = 0.0
+    return diff
+
+
 def detection_matrix(
     samples: pd.DataFrame,
     kernel: np.ndarray,
     region: Optional[str] = None,
     subsample: Optional[int] = None,
     max_dist: Optional[int] = None,
-    percentile_thresh: Optional[float] = 95.0,
+    pearson_thresh: Optional[float] = None,
+    density_thresh: Optional[float] = None,
+    mode="median",
     n_cpus: int = 4,
 ) -> Tuple[Optional[sp.csr_matrix], Optional[float]]:
     """
-    Run the detection process for a single matrix. This is abstracted from all
+    Run the detection process for a single chromosome or region. This is abstracted from all
     notions of chromosomes and genomic coordinates.
     """
     # We consider the matrix is symmetric upper (i.e. intrachromosomal)
@@ -151,28 +253,29 @@ def detection_matrix(
     else:
         map_fun = lambda x, y: [x(*args) for args in y]
 
+    # Hi-C specific preprocessing individual matrices (subsample, balance, detrend)
     samples["mat"] = map_fun(
         preprocess_hic,
         zip(samples.cool, it.repeat(min_contacts), it.repeat(region)),
     )
     print(f"{region} preprocessed", file=sys.stderr)
-
     # Return nothing if the matrix is smaller than kernel
     if np.any(np.array(samples["mat"][0].shape) <= np.array(kernel.shape)):
-        return None, None
+        return None
     # Retrieve the indices of bins which are valid in all samples (not missing
     # because of repeated sequences or low coverage)
     common_bins = pap.get_common_valid_bins(samples["mat"])
-    # Trim diagonals beyond max_dist to spare resources
+    # Trim diagonals beyond max_dist (with kernel margin for the convolution)
+    # to spare resources
     samples["mat"] = map_fun(
-        cup.diag_trim, zip(samples["mat"], it.repeat(max_dist))
+        cup.diag_trim, zip(samples["mat"], it.repeat(max_dist + max(kernel.shape)))
     )
     # Generate a missing mask from these bins
     missing_mask = cup.make_missing_mask(
         samples["mat"][0].shape,
         common_bins,
         common_bins,
-        max_dist=max_dist,
+        max_dist=max_dist + max(kernel.shape),
         sym_upper=sym_upper,
     )
     # Remove all missing values form each sample's matrix
@@ -186,6 +289,17 @@ def detection_matrix(
         ),
     )
     print(f"{region} missing bins erased", file=sys.stderr)
+
+    # Compute a density filter: regions with sufficient proportion of nonzero
+    # pixels in kernel windows, in all samples. We will use it for downstream
+    # which filter
+    if density_thresh is not None:
+        density_filter = make_density_filter(
+            samples["mat"],
+            density_thresh=density_thresh,
+            win_size=kernel.shape[0],
+            sym_upper=sym_upper,
+        )
     # Generate correlation maps for all samples using chromosight's algorithm
     corrs = map_fun(
         cud.normxcorr2,
@@ -207,43 +321,35 @@ def detection_matrix(
     # Get the union of nonzero coordinates across all samples
     total_nnz_set = pap.get_nnz_union(samples["mat"])
     # Fill zeros at these coordinates
-    samples["mat"] = samples["mat"].apply(
-        lambda cor: pap.fill_nnz(cor, total_nnz_set)
-    )
+    samples["mat"] = samples["mat"].apply(lambda cor: pap.fill_nnz(cor, total_nnz_set))
+
+    # Erase pixels where all samples are below pearson threshold
+    if pearson_thresh is not None:
+        pearson_fail = [(m.data < pearson_thresh).astype(bool) for m in samples["mat"]]
+        pearson_fail = np.bitwise_and.reduce(pearson_fail)
+        # Threshold maps using pearson correlations to reduce noisy detections
+        for i, m in enumerate(samples["mat"]):
+            m.data[pearson_fail] = 0.0
+            samples["mat"][i] = m
+
     if n_cpus > 1:
         pool.close()
-    # Compute background for each condition
-    backgrounds = samples.groupby("cond")["mat"].apply(
-        lambda g: pad.median_bg(g.reset_index(drop=True))
-    )
-    # Get the distribution of difference to background within each condition
-    within_diffs = np.hstack(
-        samples.groupby("cond")["mat"].apply(
-            lambda g: pad.reps_bg_diff(g.reset_index(drop=True))
-        )
-    )
-    # Use average difference to first background as change metric
-    diff = sp.csr_matrix(backgrounds[0].shape)
-    conditions = np.unique(samples.cond)
-    conditions = conditions[conditions != control]
-    for c in conditions:
-        diff += backgrounds[c] - backgrounds[control]
-    diff.data /= len(backgrounds) - 1
-    # Apply threshold to differences based on within-condition variations
-    try:
-        if percentile_thresh is None:
-            thresh = None
-        else:
-            thresh = np.percentile(
-                abs(within_diffs[within_diffs != 0]), percentile_thresh
-            )
-            diff.data[np.abs(diff.data) < thresh] = 0.0
-    # If there is no nonzero value (e.g. very small matrices), return nothing
-    except IndexError:
-        diff = None
-        thresh = None
 
-    return diff, thresh
+    # Use median background
+    if mode == "median":
+        diff = _median_bg_subtraction(samples, control)
+    # Use t-test on each pixel (untested, probably doesn't work well)
+    else:
+        diff = _ttest_matrix(samples, control)
+
+    # Erase pixels which do not pass the density filter in all samples
+    if density_thresh is not None:
+        diff = diff.multiply(density_filter)
+    # Remove all values beyond user-specified max_dist
+    if max_dist is not None:
+        diff = cup.diag_trim(diff, max_dist + 2)
+
+    return diff
 
 
 def change_detection_pipeline(
@@ -253,9 +359,12 @@ def change_detection_pipeline(
     bed2d_file: Optional[str] = None,
     region: Optional[Union[Iterable[str], str]] = None,
     max_dist: Optional[int] = None,
+    min_dist: Optional[int] = None,
     subsample: bool = True,
-    percentile_thresh: Optional[float] = 95.0,
+    pearson_thresh: Optional[float] = None,
+    density_thresh: float = 0.10,
     n_cpus: int = 4,
+    mode="median",
 ) -> pd.DataFrame:
     """
     Run end to end pattern change detection pipeline on input cool files. A
@@ -277,6 +386,48 @@ def change_detection_pipeline(
 
     Positive diff_scores mean the pattern intensity was increased relative to
     control (first condition).
+
+    Parameters
+    ----------
+    cool_files : Iterable of strs
+        The list of paths to cool files for the input samples.
+    conditions : Iterable of strs
+        The list of conditions matching the samples.
+    kernel : np.array of floats or str
+        Either the kernel to use as pattern as a numpy array, or the name of a
+        valid chromosight pattern.
+    bed2d_file : str or None
+        Path to a bed2D file containing a list of 2D positions. If this is
+        provided, pattern changes at these coordinates will be quantified.
+        Otherwise, they will be detected based on a threshold.
+    region : str or iterable of strs
+        Either a single UCSC format region string, or a list of multiple
+        regions. The analysis will be restricted to those regions.
+    max_dist : int or None
+        Maximum interaction distance (in basepairs) to consider in the analysis.
+        If this is not specified and a chromosight kernel was specified, the
+        default max_dist for that kernel is used. If the case of a custom kernel,
+        the whole matrix will be scanned if no max_dist is specified.
+    subsample : bool
+        Whether all input matrices should be subsampled to the same number of
+        contacts as the least covered sample.
+    pearson_thresh: float or None
+        The pearson correlation threshold to use when detecting patterns. If None,
+        the default value for the kernel is used.
+    density_thresh: float
+        The pixel density threshold to require. Low coverage windows with a
+        proportion of nonzero pixels below this value are discarded.
+    n_cpus : int
+        Number of CPU cores to allocate for parallel operations.
+    mode : str
+        Algorithm used for change detection. Can be median or stat. Median is
+        the default, stat is experimental and should only be used with at least
+        3 samples per condition.
+
+    Returns
+    -------
+    pd.DataFrame :
+        The list of reported 2D coordinates and their change intensities.
     """
     # Make sure each sample has an associated condition
     if len(cool_files) != len(conditions):
@@ -284,6 +435,12 @@ def change_detection_pipeline(
             "The lists of cool files and conditions must have the same length"
         )
 
+    if mode != "median":
+        print(
+            "Running in t-test mode, t-values will be used as diff scores.",
+            file=sys.stderr,
+        )
+
     # If a pattern name was provided, load corresponding chromosight kernel
     if isinstance(kernel, str):
         kernel_name = kernel
@@ -291,8 +448,16 @@ def change_detection_pipeline(
             kernel = getattr(ck, kernel_name)["kernels"][0]
             if max_dist is None:
                 max_dist = getattr(ck, kernel_name)["max_dist"]
+            if min_dist is None:
+                min_dist = getattr(ck, kernel_name)["min_dist"]
+            if pearson_thresh is None:
+                pearson_thresh = getattr(ck, kernel_name)["pearson"]
         except AttributeError:
             raise AttributeError(f"{kernel_name} is not a valid pattern name")
+        print(f"Loading default parameter for kernel '{kernel_name}'...")
+        print(f"pearson_thresh: {pearson_thresh}")
+        print(f"min_dist: {min_dist}")
+        print(f"max_dist: {max_dist}")
     elif isinstance(kernel, np.ndarray):
         kernel_name = "custom kernel"
     else:
@@ -300,20 +465,20 @@ def change_detection_pipeline(
             "Kernel must either be a valid chromosight pattern name, or a 2D numpy.ndarray of floats"
         )
     # Associate samples with their conditions
-    samples = pd.DataFrame(
-        {"cond": conditions, "cool": pai.get_coolers(cool_files)}
-    )
-    print(
-        f"Changes will be computed relative to condition: {samples.cond.values[0]}"
-    )
+    samples = pd.DataFrame({"cond": conditions, "cool": pai.get_coolers(cool_files)})
+    print(f"Changes will be computed relative to condition: {samples.cond.values[0]}")
     # Define each chromosome as a region, if None specified
     clr = samples.cool.values[0]
     if max_dist is not None:
         max_dist = max_dist // clr.binsize
+    if min_dist is not None:
+        min_dist = min_dist // clr.binsize
     if region is None:
         regions = clr.chroms()[:]["name"].tolist()
     elif isinstance(region, str):
-        region = [region]
+        regions = [region]
+    else:
+        regions = region
     pos_cols = [
         "chrom1",
         "start1",
@@ -334,24 +499,25 @@ def change_detection_pipeline(
     for reg in regions:
         # Subset bins to the range of interest
         bins = clr.bins().fetch(reg).reset_index(drop=True)
-        diff, thresh = detection_matrix(
+        diff = detection_matrix(
             samples,
             kernel,
             region=reg,
             subsample=subsample,
             max_dist=max_dist,
-            percentile_thresh=percentile_thresh,
+            pearson_thresh=pearson_thresh,
+            density_thresh=density_thresh,
             n_cpus=n_cpus,
+            mode=mode,
         )
-        # If the matrix was too small, skip it
-        if thresh is None and percentile_thresh is not None:
+
+        # If the matrix was too small or no difference was found, skip it
+        if diff is None or diff.nnz == 0:
             continue
         # If positions were provided, return the change value for each of them
         if bed2d_file:
             tmp_chr = reg.split(":")[0]
-            tmp_rows = (positions.chrom1 == tmp_chr) & (
-                positions.chrom2 == tmp_chr
-            )
+            tmp_rows = (positions.chrom1 == tmp_chr) & (positions.chrom2 == tmp_chr)
             # If there are no positions of interest on this chromosome, just
             # skip it
             if not np.any(tmp_rows):
@@ -360,9 +526,7 @@ def change_detection_pipeline(
             # Convert both coordinates from genomic coords to bins
             for i in [1, 2]:
                 tmp_pos["chrom"] = tmp_pos[f"chrom{i}"]
-                tmp_pos["pos"] = (
-                    tmp_pos[f"start{i}"] + tmp_pos[f"end{i}"]
-                ) // 2
+                tmp_pos["pos"] = (tmp_pos[f"start{i}"] + tmp_pos[f"end{i}"]) // 2
                 tmp_pos[f"bin{i}"] = coords_to_bins(clr, tmp_pos).astype(int)
                 # Save bin coordinates from current chromosome to the full table
                 positions.loc[tmp_rows, f"bin{i}"] = tmp_pos[f"bin{i}"]
@@ -374,7 +538,7 @@ def change_detection_pipeline(
         # Otherwise report individual spots of change using chromosight
         else:
             # Pick "foci" of changed pixels and their local maxima
-            tmp_pos, _ = cud.picker(abs(diff), thresh)
+            tmp_pos, _ = cud.pick_foci(abs(diff), 0.01, min_size=3)
             # Get genomic positions from matrix coordinates
             tmp_pos = pd.DataFrame(tmp_pos, columns=["bin1", "bin2"])
             for i in [1, 2]:
@@ -392,8 +556,22 @@ def change_detection_pipeline(
                 # Add axis' columns to  dataframe
                 tmp_pos = pd.concat([coords, tmp_pos], axis=1)
             # Retrieve diff values for each coordinate
-            tmp_pos["diff_score"] = diff[tmp_pos.bin1, tmp_pos.bin2].A1
+            try:
+                tmp_pos["diff_score"] = diff[tmp_pos.bin1, tmp_pos.bin2].A1
+            # No position found, go to next region
+            except AttributeError:
+                continue
             # Append new chromosome's rows
             positions = pd.concat([positions, tmp_pos], axis=0)
+            # For 1D patterns (e.g. borders) set diagonal positions.
+            if max_dist == 0:
+                positions[["bin1", "chrom1", "start1", "end1"]] = positions[
+                    ["bin2", "chrom2", "start2", "end2"]
+                ]
     positions = positions.loc[:, pos_cols]
+    positions = positions.loc[
+        abs(positions.bin2 - positions.bin1) >= min_dist, :
+    ].reset_index(drop=True)
+    print(min_dist)
+    print(positions)
     return positions
